with Ada.Characters.Latin_1;
package body Maps is
   procedure Put_Image (Output : in out Ada.Strings.Text_Buffers.Root_Buffer_Type'Class; M : Map) is
      First : Boolean := True;
   begin
      Output.Put ("{");
      for Key in M'Range loop
         if M (Key).Has then
            if not First then
               Output.Put (", ");
            end if;
            Output.Put ([1 => Ada.Characters.Latin_1.LF]);
            Output.Put ("   ");
            Output.Put (Key'Image);
            Output.Put (": ");
            Output.Put(Val_T'Image (M (Key).Val));
            First := False;
         end if;
      end loop;

      if not First then
         Output.Put ([1 => Ada.Characters.Latin_1.LF]);
      end if;

      Output.Put ("}");
   end Put_Image;

   function Make (M : Map_Init) return Map is
      R : Map;
   begin
      for M_KV of M loop
         R (M_KV.Key) := Has (M_KV.Val);
      end loop;

      return R;
   end Make;

   function Cursor (M : Map; First, Last : Key_T) return Map_Cursor is
      Reversed : constant Boolean := First > Last;
      Start : constant Key_T := (if not Reversed then First else Last);
      Finish : constant Key_T := (if not Reversed then Last else First);
      K : Key_T := Start;
   begin
      loop
         if M (K).Has then
            return Has (K);
         elsif K = Finish then
            return None;
         end if;

         if not Reversed then
            K := Key_T'Succ (K);
         else
            K := Key_T'Pred (K);
         end if;
      end loop;
   end Cursor;
      
   function First (M : Map) return Map_Cursor is (Cursor (M, M'First, M'Last));
   function Last (M : Map) return Map_Cursor is (Cursor (M, M'Last, M'First));

   function Next (M : Map; E : Map_Cursor) return Map_Cursor is
   begin
      if E.Has and then E.K /= M'Last then
         return Cursor (M, Key_T'Succ (E.K), M'Last);
      else
         return None;
      end if;
   end Next;

   function Previous (M : Map; E : Map_Cursor) return Map_Cursor is
   begin
      if E.Has and then E.K /= M'First then
         return Cursor (M, Key_T'Pred (E.K), M'First);
      else
         return None;
      end if;
   end Previous;

   function Has_Element (M : Map; E : Map_Cursor) return Boolean is
   begin
      return E.Has;
   end Has_Element;

   function Element (M : Map; E : Map_Cursor) return Map_Element is
   begin
      return (Key => E.K, Val => M (E.K).Val);
   end Element;
end Maps;
