with Ada.Strings.Text_Buffers;
generic
   type Key_T is (<>);
   type Val_T is private;
package Maps is

   type Map_Element is record
      Key : Key_T;
      Val : Val_T;
   end record;

   type Map is private
      with Put_Image => Put_Image,
         Iterable => (
            First => First,
            Next => Next,
            Has_Element => Has_Element,
            Element => Element,
            Last => Last,
            Previous => Previous
         );
   type Map_Cursor (<>) is private;
   type Map_Init is array (Positive range <>) of Map_Element;
   function Make (M : Map_Init) return Map;
   procedure Put_Image (Output : in out Ada.Strings.Text_Buffers.Root_Buffer_Type'Class; M : Map);

   function First (M : Map) return Map_Cursor;
   function Next (M : Map; E : Map_Cursor) return Map_Cursor;
   function Has_Element (M : Map; E : Map_Cursor) return Boolean;
   function Element (M : Map; E : Map_Cursor) return Map_Element
      with Pre => Has_Element (M, E);
   function Last (M : Map) return Map_Cursor;
   function Previous (M : Map; E : Map_Cursor) return Map_Cursor;

private
   type Maybe_Val (Has : Boolean := False) is record
      case Has is
      when True =>
         Val : Val_T;
      when False =>
         null;
      end case;
   end record;
   function Has (Val : Val_T) return Maybe_Val is (Has => True, Val => Val);
   function None return Maybe_Val is (Has => False);

   type Map_Cursor (Has : Boolean := False) is record
      case Has is
      when True =>
         K : Key_T;
      when False =>
         null;
      end case;
   end record;
   function Has (K : Key_T) return Map_Cursor is (Has => True, K => K);
   function None return Map_Cursor is (Has => False);

   type Map is array (Key_T) of Maybe_Val;
end Maps;
